/*
 * Copyright (C) 2009 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.android.marvin.talkback.formatter;

import android.content.res.Resources;
import android.os.Build;
import android.os.Bundle;
import android.text.TextUtils;
import android.util.Log;
import android.view.accessibility.AccessibilityEvent;
import android.view.accessibility.AccessibilityNodeInfo;

import jp.kyocera.kcfp.util.KCfpSoftkeyGuide;

import com.google.android.marvin.talkback.R;
import com.google.android.marvin.talkback.SpeechCleanupUtils;
import com.google.android.marvin.talkback.SpeechController;
import com.google.android.marvin.talkback.TalkBackService;
import com.google.android.marvin.talkback.Utterance;
import com.google.android.marvin.talkback.formatter.EventSpeechRule.AccessibilityEventFilter;
import com.google.android.marvin.talkback.formatter.EventSpeechRule.AccessibilityEventFormatter;
import com.googlecode.eyesfree.utils.AccessibilityEventUtils;
import com.googlecode.eyesfree.utils.LogUtils;
import com.googlecode.eyesfree.utils.StringBuilderUtils;
import vendor.kyocera.hardware.kcjprop.V1_0.IKcjprop;
import android.os.RemoteException;
import org.w3c.dom.Document;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

/**
 * This class is a {@link EventSpeechRule} processor responsible for loading
 * from speech strategy XML files sets of {@link EventSpeechRule}s used for
 * processing {@link AccessibilityEvent}s such that utterances are generated.
 * <p>
 * Speech strategies can be registered for handling events from a given package
 * or their rules to be appended to the default speech rules which are examined
 * as fall-back if no package specific ones have matched the event. The rules
 * are processed in the order they are defined and in case a rule is
 * successfully applied i.e. an utterance is formatted, processing stops. In
 * other words, the first applicable speech rule wins.
 * </p>
 *
 * @author svetoslavganov@google.com (Svetoslav Ganov)
 */
public class EventSpeechRuleProcessor {
    private static final String TAG = "EventSpeechRuleProcessor";

    /**
     * Indicates the result of filtering and formatting an
     * {@link AccessibilityEvent} through the processor.
     */
    private enum RuleProcessorResult {
        /**
         * Result indicating that an {@link EventSpeechRule} matched an
         * {@link AccessibilityEvent} and formatted an {@link Utterance}
         */
        FORMATTED,

        /**
         * Result indicating that no {@link EventSpeechRule}'s
         * {@link AccessibilityEventFilter} matched an
         * {@link AccessibilityEvent}
         */
        NOT_MATCHED,

        /**
         * Result indicating that an {@link EventSpeechRule}'s
         * {@link AccessibilityEventFilter} matched an
         * {@link AccessibilityEvent}, but its
         * {@link AccessibilityEventFormatter} indicated the event should be
         * dropped from the processor.
         */
        REJECTED
    }

    /**
     * Constant used for storing all speech rules that either do not define a
     * filter package or have custom filters.
     */
    private static final String UNDEFINED_PACKAGE_NAME = "undefined_package_name";

    private static final long IME_WAIT_TIME = 250; /* ms */

    /** Context for accessing resources. */
    private final TalkBackService mContext;

    /** Mapping from package name to speech rules for that package. */
    private final Map<String, List<EventSpeechRule>>
            mPackageNameToSpeechRulesMap = new HashMap<String, List<EventSpeechRule>>();

    private final Object mLock = new Object();

    /** A lazily-constructed shared instance of a document builder. */
    private DocumentBuilder mDocumentBuilder;

    /** keep the status which the keyboard is hidden or not. */
    private boolean mIsKeyboardHidden;

    private static IKcjprop mServer = null;
    /**
     * Creates a new instance of a speech rule processor.
     *
     * @param context The service context.
     */
    public EventSpeechRuleProcessor(TalkBackService context) {
        mContext = context;
    }

    /**
     * Processes an {@code event} by sequentially trying to apply all
     * {@link EventSpeechRule}s maintained by this processor in the order they
     * are defined for the package source of the event. If no package specific
     * rules exist the default speech rules are examined in the same manner. If
     * a rule is successfully applied the result is used to populate an
     * {@code utterance}. In other words, the first matching rule wins.
     *
     * @return {@code true} if the event was processed, {@code false} otherwise.
     */
    public boolean processEvent(AccessibilityEvent event, Utterance utterance) {
        synchronized (mLock) {
            int eventType = event.getEventType();

            if (eventType == AccessibilityEvent.TYPE_ANNOUNCEMENT) {
                CharSequence announcementText = AccessibilityEventUtils.getEventAggregateText(event);
                Matcher matcher = TalkBackService.TYPE_ANNOUNCEMENT_READOUT_PATTERN.matcher(announcementText);

                final int startLength = TalkBackService.START_TAG_READOUT_ANNOUNCEMENT.length();
                final int endLength = TalkBackService.END_TAG_READOUT_ANNOUNCEMENT.length();

                while (matcher.find()) {
                    String readoutTextWithTag = matcher.group();
                    String readoutText = readoutTextWithTag.substring(startLength,
                            (readoutTextWithTag.length() - endLength));
                    if (!TextUtils.isEmpty(readoutText)) {
                        appendTextToUtterance(readoutText, utterance);
                    }
                }

                if (utterance.getSpoken().isEmpty()) {
                    Matcher addTagMatcher = TalkBackService.TYPE_ANNOUNCEMENT_READOUT_ADD_PATTERN.matcher(announcementText);
                    final int addTagStartLength = TalkBackService.START_TAG_READOUT_ANNOUNCEMENT_ADD.length();
                    final int addTagEndLength = TalkBackService.END_TAG_READOUT_ANNOUNCEMENT_ADD.length();

                    while (addTagMatcher.find()) {
                        String readoutTextWithTag = addTagMatcher.group();
                        String readoutText = readoutTextWithTag.substring(addTagStartLength,
                                (readoutTextWithTag.length() - addTagEndLength));
                        if (!TextUtils.isEmpty(readoutText)) {
                            appendTextToUtterance(readoutText, utterance);
                        }
                    }

                    if (!utterance.getSpoken().isEmpty()) {
                        utterance.getMetadata().putInt(Utterance.KEY_METADATA_QUEUING, SpeechController.QUEUE_MODE_QUEUE);
                    }
                }
            } else if (event.getOriginalReadoutText() != null
                    && event.isReadoutOriginalTextWithoutSoftkey()) {
                String readoutText = event.getOriginalReadoutText();
                appendTextToUtterance(readoutText, utterance);
            } else {
                AccessibilityNodeInfo parent = event.getSource();

                if(eventType == AccessibilityEvent.TYPE_WINDOW_STATE_CHANGED &&
                    parent != null &&
                    parent.getPackageName() != null &&
                    "com.kyocera.mediumkey".equals(parent.getPackageName().toString()) &&
                    event.getClassName() != null &&
                    "com.kyocera.mediumkey.settings.TextEntrySettingsActivity".equals(event.getClassName().toString())){
                    try{
                        Thread.sleep(IME_WAIT_TIME);
                    }catch(Exception e){}
                    parent.refresh();
                }
                if (parent == null) {
                    return false;
                }

                while (parent.getParent() != null) {
                    parent = parent.getParent();
                }

                if ("eng".equals(Build.TYPE) || "userdebug".equals(Build.TYPE)) {
                    Log.v(TAG, "[KcReadout] AccessibilityNodeInfo of root viewgroup : " + parent);
                }

                KcReadoutBuilder kcReadoutBuilder = new KcReadoutBuilder();
                if (event.getText() != null && event.getText().size() > 0) {
                    kcReadoutBuilder.setAccessibilityEventText(event.getText().get(0));
                }

                mIsKeyboardHidden = isKeyboardHidden();
                for (int i=0;i<parent.getChildCount();i++) {
                    setChildInfoToKcReadoutBuilder(kcReadoutBuilder, parent, i);
                }

                CharSequence softkeyText = KCfpSoftkeyGuide.getRealReadoutText();
                if (softkeyText != null && !TextUtils.isEmpty(softkeyText) && mIsKeyboardHidden) {
                    kcReadoutBuilder.setSoftkeyFullText(softkeyText);
                }

                kcReadoutBuilder.buildKcReadout(event, utterance);
            }

            return !TextUtils.isEmpty(StringBuilderUtils.getAggregateText(utterance.getSpoken()));
        }
    }

    /**
     * Loads a speech strategy from a given <code>resourceId</code> to handle events from
     * the specified <code>targetPackage</code> and use the resources from a given <code>context
     * </code>. If the target package is <code>null</code> the rules of the loaded
     * speech strategy are appended to the default speech rules. While for
     * loading of resources is used the provided context instance, for loading
     * plug-in classes (custom Filters and Formatters) the <code>publicSourceDir</code> which
     * specifies the location of the APK that defines them is used to enabled
     * using the TalkBack {@link ClassLoader}.
     */
    public void addSpeechStrategy(int resourceId) {
        final Resources res = mContext.getResources();
        final String speechStrategy = res.getResourceName(resourceId);
        final InputStream inputStream = res.openRawResource(resourceId);
        final Document document = parseSpeechStrategy(inputStream);
        final ArrayList<EventSpeechRule> speechRules = EventSpeechRule.createSpeechRules(
                mContext, document);

        final int added = addSpeechStrategy(speechRules);

        LogUtils.log(EventSpeechRuleProcessor.class, Log.INFO, "%d speech rules appended from: %s",
                added, speechStrategy);
    }

    /**
     * Loads speech rules from a list.
     *
     * @return The number of rules that were loaded successfully.
     */
    public int addSpeechStrategy(Iterable<EventSpeechRule> speechRules) {
        int count = 0;

        synchronized (mPackageNameToSpeechRulesMap) {
            for (EventSpeechRule speechRule : speechRules) {
                if (addSpeechRuleLocked(speechRule)) {
                    count++;
                }
            }
        }

        return count;
    }

    /**
     * Adds a <code>speechRule</code>.
     */
    private boolean addSpeechRuleLocked(EventSpeechRule speechRule) {
        final String packageName = speechRule.getPackageName();

        List<EventSpeechRule> packageSpeechRules = mPackageNameToSpeechRulesMap.get(packageName);

        if (packageSpeechRules == null) {
            packageSpeechRules = new LinkedList<EventSpeechRule>();
            mPackageNameToSpeechRulesMap.put(packageName, packageSpeechRules);
        }

        return packageSpeechRules.add(speechRule);
    }

    /**
     * Processes an {@code event} by sequentially trying to apply all
     * {@code speechRules} in the order they are defined. If a rule is
     * successfully applied the result is used to populate an {@code Utterance}.
     *
     * @return a {@link RuleProcessorResult} corresponding to how the event was
     *         processed.
     */
    private RuleProcessorResult processEvent(
            List<EventSpeechRule> speechRules, AccessibilityEvent event, Utterance utterance) {
        for (EventSpeechRule speechRule : speechRules) {
            // We should never crash because of a bug in speech rules.
            try {
                if (speechRule.applyFilter(event)) {
                    if (speechRule.applyFormatter(event, utterance)) {
                        LogUtils.log(EventSpeechRuleProcessor.class, Log.VERBOSE,
                                "Processed event using rule:\n%s", speechRule);
                        return RuleProcessorResult.FORMATTED;
                    } else {
                        LogUtils.log(EventSpeechRuleProcessor.class, Log.VERBOSE,
                                "The \"%s\" filter accepted the event, but the \"%s\" "
                                        + "formatter indicated the event should be dropped.",
                                speechRule.getFilter().getClass().getSimpleName(),
                                speechRule.getFormatter().getClass().getSimpleName());
                        return RuleProcessorResult.REJECTED;
                    }
                }
            } catch (Exception e) {
                LogUtils.log(EventSpeechRuleProcessor.class, Log.ERROR,
                        "Error while processing rule:\n%s", speechRule);
                e.printStackTrace();
            }
        }

        return RuleProcessorResult.NOT_MATCHED;
    }

    /**
     * Parses a speech strategy XML file specified by <code>resourceId</code> and returns
     * a <code>document</code>. If an error occurs during the parsing, it is logged and
     * <code>null</code> is returned.
     *
     * @param inputStream An {@link InputStream} to the speech strategy XML
     *            file.
     * @return The parsed {@link Document} or <code>null</code> if an error
     *         occurred.
     */
    private Document parseSpeechStrategy(InputStream inputStream) {
        try {
            final DocumentBuilder builder = getDocumentBuilder();
            return builder.parse(inputStream);
        } catch (Exception e) {
            LogUtils.log(EventSpeechRuleProcessor.class, Log.ERROR,
                    "Could not open speechstrategy xml file\n%s", e.toString());
        }

        return null;
    }

    /**
     * @return A lazily-constructed shared instance of a document builder.
     * @throws ParserConfigurationException
     */
    private DocumentBuilder getDocumentBuilder() throws ParserConfigurationException {
        if (mDocumentBuilder == null) {
            mDocumentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        }

        return mDocumentBuilder;
    }

    private void appendTextToUtterance(CharSequence text, Utterance utterance) {
        CharSequence readoutText = SpeechCleanupUtils.replaceSpecificWords(mContext, text);
        utterance.addSpoken(readoutText);
    }

    private void setChildInfoToKcReadoutBuilder(KcReadoutBuilder kcReadoutBuilder,
            AccessibilityNodeInfo parent, int childNumber) {
        AccessibilityNodeInfo child = parent.getChild(childNumber);
        if ("eng".equals(Build.TYPE) || "userdebug".equals(Build.TYPE)) {
            Log.v(TAG, "[KcReadout] AccessibilityNodeInfo of child view " + childNumber + " : " + child);
        }

        if (child == null) {
            return;
        }

        CharSequence nodeText = child.getContentDescription() != null ?
            child.getContentDescription() : child.getText();

        if ("android.widget.Spinner".equals(child.getClassName()) && child.isFocused() &&
                TextUtils.isEmpty(nodeText)) {
            CharSequence spinnerReadoutText = getSpinnerReadoutText(kcReadoutBuilder, child);
            if (spinnerReadoutText != null) {
                nodeText = spinnerReadoutText;
            }
        }

        if (child.isVisibleToUser()) {
            if (child.isFocused()) {
                kcReadoutBuilder.addFocusedText(nodeText);
            } else if (child.isSelected()) {
                kcReadoutBuilder.addSelectedText(nodeText);
            } else {
                kcReadoutBuilder.addOtherText(nodeText, child.getText());
            }
        }

        int childCount = child.getChildCount();
        if (childCount == 0) {
            return;
        }

        int currentChild = 0;
        while (currentChild < childCount) {
            setChildInfoToKcReadoutBuilder(kcReadoutBuilder, child, currentChild);
            currentChild++;
        }
    }

    private CharSequence getSpinnerReadoutText(KcReadoutBuilder kcReadoutBuilder, AccessibilityNodeInfo parent) {
        int childCount = parent.getChildCount();
        if (childCount != 1) {
            return null;
        }

        AccessibilityNodeInfo child = parent.getChild(0);
        if (child == null) {
            return null;
        }

        if (!"android.widget.TextView".equals(child.getClassName()) || child.isFocused()) {
            return null;
        }

        CharSequence readoutText = kcReadoutBuilder.getAccessibilityEventText();
        if (TextUtils.isEmpty(readoutText) || !readoutText.equals(child.getText())) {
            return null;
        }

        return readoutText;

    }

    String mKeyboardStatus = "0";

    private boolean isKeyboardHidden() {
        try {
            if (mServer == null) {
                mServer = IKcjprop.getService(true /* retry */);
            }
        } catch (RemoteException e) {
            Log.e(TAG, "RemoteException:" + e);
        }
        try {
            mServer.get("rw_mediumkey_showing_now_flag", new IKcjprop.getCallback() {
                @Override
                public void onValues(int ret, String outData) {
                    if (ret != 0) {
                        Log.e(TAG, "error:" + ret);
                    } else {
                        mKeyboardStatus = outData;
                    }
                }
            });
        } catch (RemoteException e) {
            Log.e(TAG, "RemoteException:" + e);
        }
        return ("0".equals(mKeyboardStatus) || "3".equals(mKeyboardStatus));
    }

    private class KcReadoutBuilder {
        private CharSequence mAccessibilityEventText;
        private CharSequence mTitleText;
        private CharSequence mLeftSoftkeyText;
        private CharSequence mCenterSoftkeyText;
        private CharSequence mRightSoftkeyText;
        private CharSequence mSoftkeyFullText;
        private ArrayList<CharSequence> mFocusedText;
        private ArrayList<CharSequence> mSelectedText;
        private ArrayList<CharSequence> mOtherText;

        public KcReadoutBuilder() {
            mTitleText = null;
            mLeftSoftkeyText = null;
            mCenterSoftkeyText = null;
            mRightSoftkeyText = null;
            mSoftkeyFullText = null;
            mFocusedText = new ArrayList<CharSequence>();
            mSelectedText = new ArrayList<CharSequence>();
            mOtherText = new ArrayList<CharSequence>();
        }

        public void setAccessibilityEventText(CharSequence text) {
            mAccessibilityEventText = text;
        }

        public void setSoftkeyText(int type, CharSequence text) {
            // switch (type) {
            //     case AccessibilityNodeInfo.SOFTKEY_LEFT:
            //         mLeftSoftkeyText = text;
            //         break;
            //     case AccessibilityNodeInfo.SOFTKEY_CENTER:
            //         mCenterSoftkeyText = text;
            //         break;
            //     case AccessibilityNodeInfo.SOFTKEY_RIGHT:
            //         mRightSoftkeyText = text;
            //         break;
            //     default:
            //         /* Not run this process. */
            // }
        }

        public void setSoftkeyFullText(CharSequence text) {
            mSoftkeyFullText = text;
        }

        public CharSequence getAccessibilityEventText() {
            return mAccessibilityEventText;
        }

        public void addFocusedText(CharSequence text) {
            if (TextUtils.isEmpty(text)) {
                return;
            }
            mFocusedText.add(text);
        }

        public void addSelectedText(CharSequence text) {
            if (TextUtils.isEmpty(text)) {
                return;
            }
            mSelectedText.add(text);
        }

        public void addOtherText(CharSequence text, CharSequence displayText) {
            if (TextUtils.isEmpty(text)) {
                return;
            }

            boolean isTitleText = mOtherText.size() == 0 &&
                mAccessibilityEventText != null &&
                (mAccessibilityEventText.toString().equals(text.toString()) ||
                 (displayText != null &&
                  mAccessibilityEventText.toString().equals(displayText.toString())));

            if (isTitleText) {
                mTitleText = text;
            } else {
                mOtherText.add(text);
            }
        }

        public void buildKcReadout(AccessibilityEvent event, Utterance utterance) {
            if (event.getOriginalReadoutText() != null) {
                String readoutText = event.getOriginalReadoutText();
                appendTextToUtterance(readoutText, utterance);
            } else {
                switch (event.getEventType()) {
                    case AccessibilityEvent.TYPE_WINDOW_STATE_CHANGED:
                        if (!TextUtils.isEmpty(mTitleText)) {
                            appendTextToUtterance(mTitleText, utterance);
                        }
                        if (mFocusedText.size() > 0) {
                            for(int i=0;i<mFocusedText.size();i++) {
                                appendTextToUtterance(mFocusedText.get(i), utterance);
                            }
                        } else if (mSelectedText.size() > 0) {
                            for(int i=0;i<mSelectedText.size();i++) {
                                appendTextToUtterance(mSelectedText.get(i), utterance);
                            }
                        } else if (mOtherText.size() == 1) {
                            appendTextToUtterance(mOtherText.get(0), utterance);
                        } else {
                            /* don't readout body text */
                        }
                        break;
                    case AccessibilityEvent.TYPE_VIEW_FOCUSED:
                        if (mFocusedText.size() > 0) {
                            for(int i=0;i<mFocusedText.size();i++) {
                                appendTextToUtterance(mFocusedText.get(i), utterance);
                            }
                        }
                        break;
                    case AccessibilityEvent.TYPE_VIEW_SELECTED:
                        if (mSelectedText.size() > 0) {
                            for(int i=0;i<mSelectedText.size();i++) {
                                appendTextToUtterance(mSelectedText.get(i), utterance);
                            }
                        }
                        break;
                    default:
                        /* Not run this process. */
                }
            }

            if (!TextUtils.isEmpty(mSoftkeyFullText)) {
                appendTextToUtterance(mSoftkeyFullText, utterance);
            }
        }
    }
}
